<!DOCTYPE html>
<html lang="en">
<head>

  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
  <meta name="generator" content="{{ generator }}" />
  <meta name="description" content="Automatically generated population shift report" />
  <meta name="url" content="https://github.com/ing-bank/popmon" />

  <title>{{ title | striptags }}</title>

  {%- if online_report %}
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.3.1/dist/css/bootstrap.min.css">
  {%- endif -%}

  <style>
    {%- if not online_report -%}
        {%- include 'assets/css/bootstrap.min.css' -%}
    {%- endif -%}
    {%- include 'assets/css/custom-style.css' -%}
  </style>

</head>
<body id="page-top">
    {%- include 'modal-popup.html' -%}
    <script>
    var rendered = {};
    var layouts = {};
    var render_func = {};
    </script>
    {%- include 'header.html' -%}
    {{ sections }}
    {%- include 'footer.html' -%}
</body>
</html>